/*****************************************************************************
 * os2_thread.c : OS/2 back-end for LibVLC
 *****************************************************************************
 * Copyright (C) 2010 KO Myung-Hun
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>

#include "libvlc.h"
#include <stdarg.h>
#include <assert.h>
#include <limits.h>
#include <errno.h>

static vlc_threadvar_t cancel_key;

/**
 * Per-thread cancellation data
 */
typedef struct vlc_cancel_t
{
    vlc_cleanup_t *cleaners;
    bool           killable;
    bool           killed;
    HEV            hevCancel;
    HEV            hevDone;
} vlc_cancel_t;

# define VLC_CANCEL_INIT { NULL, true, false, 0, 0 }

static void vlc_cancel_self (PVOID dummy);

static APIRET DosWaitEventSemEx( HEV hev, ULONG ulTimeout, BOOL fCancelable )
{
    HMUX      hmux;
    SEMRECORD asr[ 2 ];
    ULONG     ulUser;
    ULONG     rc;

    vlc_cancel_t *nfo = vlc_threadvar_get( cancel_key );
    if( nfo == NULL || !fCancelable )
    {
        /* Main thread - cannot be cancelled anyway */
        return DosWaitEventSem( hev, ulTimeout );
    }

    asr[ 0 ].hsemCur = ( HSEM )hev;
    asr[ 0 ].ulUser  = 0;
    asr[ 1 ].hsemCur = ( HSEM )nfo->hevCancel;
    asr[ 1 ].ulUser  = 0xFFFF;
    DosCreateMuxWaitSem( NULL, &hmux, 2, asr, DCMW_WAIT_ANY );
    rc = DosWaitMuxWaitSem( hmux, ulTimeout, &ulUser );
    DosCloseMuxWaitSem( hmux );
    if( rc )
        return rc;

    if( ulUser == 0xFFFF )
    {
        vlc_cancel_self( NULL );
        return ERROR_INTERRUPT;
    }

    return NO_ERROR;
}

vlc_mutex_t super_mutex;
vlc_cond_t  super_variable;

int _CRT_init(void);
void _CRT_term(void);

unsigned long _System _DLL_InitTerm(unsigned long hmod, unsigned long flag)
{
    switch (flag)
    {
        case 0 :    /* Initialization */
            if(_CRT_init() == -1)
                return 0;

            vlc_mutex_init (&super_mutex);
            vlc_cond_init (&super_variable);
            vlc_threadvar_create (&cancel_key, free);

            return 1;

        case 1 :    /* Termination */
            vlc_threadvar_delete( &cancel_key );
            vlc_cond_destroy (&super_variable);
            vlc_mutex_destroy (&super_mutex);

            _CRT_term();

            return 1;
    }

    return 0;   /* Failed */
}

/*** Mutexes ***/
void vlc_mutex_init( vlc_mutex_t *p_mutex )
{
    /* This creates a recursive mutex. This is OK as fast mutexes have
     * no defined behavior in case of recursive locking. */
    DosCreateMutexSem( NULL, &p_mutex->hmtx, 0, FALSE );
    p_mutex->dynamic = true;
}

void vlc_mutex_init_recursive( vlc_mutex_t *p_mutex )
{
    DosCreateMutexSem( NULL, &p_mutex->hmtx, 0, FALSE );
    p_mutex->dynamic = true;
}


void vlc_mutex_destroy (vlc_mutex_t *p_mutex)
{
    assert (p_mutex->dynamic);
    DosCloseMutexSem( p_mutex->hmtx );
}

void vlc_mutex_lock (vlc_mutex_t *p_mutex)
{
    if (!p_mutex->dynamic)
    {   /* static mutexes */
        int canc = vlc_savecancel ();
        assert (p_mutex != &super_mutex); /* this one cannot be static */

        vlc_mutex_lock (&super_mutex);
        while (p_mutex->locked)
        {
            p_mutex->contention++;
            vlc_cond_wait (&super_variable, &super_mutex);
            p_mutex->contention--;
        }
        p_mutex->locked = true;
        vlc_mutex_unlock (&super_mutex);
        vlc_restorecancel (canc);
        return;
    }

    DosRequestMutexSem(p_mutex->hmtx, SEM_INDEFINITE_WAIT);
}

int vlc_mutex_trylock (vlc_mutex_t *p_mutex)
{
    if (!p_mutex->dynamic)
    {   /* static mutexes */
        int ret = EBUSY;

        assert (p_mutex != &super_mutex); /* this one cannot be static */
        vlc_mutex_lock (&super_mutex);
        if (!p_mutex->locked)
        {
            p_mutex->locked = true;
            ret = 0;
        }
        vlc_mutex_unlock (&super_mutex);
        return ret;
    }

    return DosRequestMutexSem( p_mutex->hmtx, 0 ) ? EBUSY : 0;
}

void vlc_mutex_unlock (vlc_mutex_t *p_mutex)
{
    if (!p_mutex->dynamic)
    {   /* static mutexes */
        assert (p_mutex != &super_mutex); /* this one cannot be static */

        vlc_mutex_lock (&super_mutex);
        assert (p_mutex->locked);
        p_mutex->locked = false;
        if (p_mutex->contention)
            vlc_cond_broadcast (&super_variable);
        vlc_mutex_unlock (&super_mutex);
        return;
    }

    DosReleaseMutexSem( p_mutex->hmtx );
}

/*** Condition variables ***/
enum
{
    CLOCK_MONOTONIC,
    CLOCK_REALTIME,
};

static void vlc_cond_init_common (vlc_cond_t *p_condvar, unsigned clock)
{
    /* Create a manual-reset event (manual reset is needed for broadcast). */
    if (DosCreateEventSem (NULL, &p_condvar->hev, 0, FALSE))
        abort();
    p_condvar->clock = clock;
}

void vlc_cond_init (vlc_cond_t *p_condvar)
{
    vlc_cond_init_common (p_condvar, CLOCK_MONOTONIC);
}

void vlc_cond_init_daytime (vlc_cond_t *p_condvar)
{
    vlc_cond_init_common (p_condvar, CLOCK_REALTIME);
}

void vlc_cond_destroy (vlc_cond_t *p_condvar)
{
    DosCloseEventSem( p_condvar->hev );
}

void vlc_cond_signal (vlc_cond_t *p_condvar)
{
    /* NOTE: This will cause a broadcast, that is wrong.
     * This will also wake up the next waiting thread if no threads are yet
     * waiting, which is also wrong. However both of these issues are allowed
     * by the provision for spurious wakeups. Better have too many wakeups
     * than too few (= deadlocks). */
    DosPostEventSem( p_condvar->hev );
}

void vlc_cond_broadcast (vlc_cond_t *p_condvar)
{
    DosPostEventSem( p_condvar->hev );
}

void vlc_cond_wait (vlc_cond_t *p_condvar, vlc_mutex_t *p_mutex)
{
    ULONG ulPost;
    ULONG rc;

    assert (p_mutex->dynamic); /* TODO */
    do
    {
        vlc_testcancel();

        DosReleaseMutexSem( p_mutex->hmtx );
        rc = DosWaitEventSemEx( p_condvar->hev, SEM_INDEFINITE_WAIT, TRUE );
        DosRequestMutexSem( p_mutex->hmtx, SEM_INDEFINITE_WAIT );
    } while( rc == ERROR_INTERRUPT );

    DosResetEventSem( p_condvar->hev, &ulPost );
}

int vlc_cond_timedwait (vlc_cond_t *p_condvar, vlc_mutex_t *p_mutex,
                        mtime_t deadline)
{
    ULONG   ulTimeout;
    ULONG   ulPost;
    ULONG   rc;

    assert (p_mutex->dynamic); /* TODO */
    do
    {
        vlc_testcancel();

        mtime_t total;
        switch (p_condvar->clock)
        {
            case CLOCK_MONOTONIC:
                total = mdate();
                break;
            case CLOCK_REALTIME: /* FIXME? sub-second precision */
                total = CLOCK_FREQ * time (NULL);
                break;
            default:
                assert (0);
        }
        total = (deadline - total) / 1000;
        if( total < 0 )
            total = 0;

        ulTimeout = ( total > 0x7fffffff ) ? 0x7fffffff : total;

        DosReleaseMutexSem( p_mutex->hmtx );
        rc = DosWaitEventSemEx( p_condvar->hev, ulTimeout, TRUE );
        DosRequestMutexSem( p_mutex->hmtx, SEM_INDEFINITE_WAIT );
    } while( rc == ERROR_INTERRUPT );

    DosResetEventSem( p_condvar->hev, &ulPost );

    return rc ? ETIMEDOUT : 0;
}

/*** Semaphore ***/
void vlc_sem_init (vlc_sem_t *sem, unsigned value)
{
    if (DosCreateEventSem(NULL, &sem->hev, 0, value > 0 ? TRUE : FALSE))
        abort ();

    if (DosCreateMutexSem(NULL, &sem->wait_mutex, 0, FALSE))
        abort ();

    if (DosCreateMutexSem(NULL, &sem->count_mutex, 0, FALSE))
        abort ();

    sem->count     = value;
    sem->max_count = 0x7FFFFFFF;
}

void vlc_sem_destroy (vlc_sem_t *sem)
{
    DosCloseEventSem (sem->hev);
    DosCloseMutexSem (sem->wait_mutex);
    DosCloseMutexSem (sem->count_mutex);
}

int vlc_sem_post (vlc_sem_t *sem)
{
    DosRequestMutexSem(sem->count_mutex, SEM_INDEFINITE_WAIT);

    if (sem->count < sem->max_count)
    {
        sem->count++;
        DosPostEventSem(sem->hev);
    }

    DosReleaseMutexSem(sem->count_mutex);

    return 0; /* FIXME */
}

void vlc_sem_wait (vlc_sem_t *sem)
{
    ULONG rc;

    do
    {
        vlc_testcancel ();

        DosRequestMutexSem(sem->wait_mutex, SEM_INDEFINITE_WAIT);

        rc = DosWaitEventSemEx (sem->hev, SEM_INDEFINITE_WAIT, TRUE );

        if (!rc)
        {
            DosRequestMutexSem(sem->count_mutex, SEM_INDEFINITE_WAIT);

            sem->count--;
            if (sem->count == 0)
            {
                ULONG ulPost;

                DosResetEventSem(sem->hev, &ulPost);
            }

            DosReleaseMutexSem(sem->count_mutex);
        }

        DosReleaseMutexSem(sem->wait_mutex);
    } while (rc == ERROR_INTERRUPT);
}

/*** Read/write locks */
/* SRW (Slim Read Write) locks are available in Vista+ only */
void vlc_rwlock_init (vlc_rwlock_t *lock)
{
    vlc_mutex_init (&lock->mutex);
    vlc_cond_init (&lock->read_wait);
    vlc_cond_init (&lock->write_wait);
    lock->readers = 0; /* active readers */
    lock->writers = 0; /* waiting or active writers */
    lock->writer = 0; /* ID of active writer */
}

/**
 * Destroys an initialized unused read/write lock.
 */
void vlc_rwlock_destroy (vlc_rwlock_t *lock)
{
    vlc_cond_destroy (&lock->read_wait);
    vlc_cond_destroy (&lock->write_wait);
    vlc_mutex_destroy (&lock->mutex);
}

/**
 * Acquires a read/write lock for reading. Recursion is allowed.
 */
void vlc_rwlock_rdlock (vlc_rwlock_t *lock)
{
    vlc_mutex_lock (&lock->mutex);
    while (lock->writer != 0)
        vlc_cond_wait (&lock->read_wait, &lock->mutex);
    if (lock->readers == ULONG_MAX)
        abort ();
    lock->readers++;
    vlc_mutex_unlock (&lock->mutex);
}

/**
 * Acquires a read/write lock for writing. Recursion is not allowed.
 */
void vlc_rwlock_wrlock (vlc_rwlock_t *lock)
{
    vlc_mutex_lock (&lock->mutex);
    if (lock->writers == ULONG_MAX)
        abort ();
    lock->writers++;
    while ((lock->readers > 0) || (lock->writer != 0))
        vlc_cond_wait (&lock->write_wait, &lock->mutex);
    lock->writers--;
    lock->writer = _gettid();
    vlc_mutex_unlock (&lock->mutex);
}

/**
 * Releases a read/write lock.
 */
void vlc_rwlock_unlock (vlc_rwlock_t *lock)
{
    vlc_mutex_lock (&lock->mutex);
    if (lock->readers > 0)
        lock->readers--; /* Read unlock */
    else
        lock->writer = 0; /* Write unlock */

    if (lock->writers > 0)
    {
        if (lock->readers == 0)
            vlc_cond_signal (&lock->write_wait);
    }
    else
        vlc_cond_broadcast (&lock->read_wait);
    vlc_mutex_unlock (&lock->mutex);
}

/*** Thread-specific variables (TLS) ***/
int vlc_threadvar_create (vlc_threadvar_t *p_tls, void (*destr) (void *))
{
#warning FIXME: use destr() callback and stop leaking!
    VLC_UNUSED( destr );

    ULONG rc;

    rc = DosAllocThreadLocalMemory( 1, &p_tls->p );
    return rc ? ENOMEM : 0;
}

void vlc_threadvar_delete (vlc_threadvar_t *p_tls)
{
    DosFreeThreadLocalMemory( p_tls->p );
}

/**
 * Sets a thread-local variable.
 * @param key thread-local variable key (created with vlc_threadvar_create())
 * @param value new value for the variable for the calling thread
 * @return 0 on success, a system error code otherwise.
 */
int vlc_threadvar_set (vlc_threadvar_t key, void *value)
{
    *key.p = ( ULONG )value;
    return 0;
}

/**
 * Gets the value of a thread-local variable for the calling thread.
 * This function cannot fail.
 * @return the value associated with the given variable for the calling
 * or NULL if there is no value.
 */
void *vlc_threadvar_get (vlc_threadvar_t key)
{
    return ( void * )*key.p;
}


/*** Threads ***/
void vlc_threads_setup (libvlc_int_t *p_libvlc)
{
    (void) p_libvlc;
}

struct vlc_entry_data
{
    void *  (*func) (void *);
    void *    data;
    vlc_sem_t ready;
    HEV       hevCancel;
    HEV       hevDone;
};

static void vlc_entry( void *p )
{
    struct vlc_entry_data *entry = p;
    vlc_cancel_t cancel_data = VLC_CANCEL_INIT;
    void *(*func) (void *) = entry->func;
    void *data = entry->data;

    cancel_data.hevCancel = entry->hevCancel;
    cancel_data.hevDone   = entry->hevDone;

    vlc_threadvar_set( cancel_key, &cancel_data );

    vlc_sem_post (&entry->ready);
    func (data);

    DosPostEventSem( cancel_data.hevDone );
}


int vlc_clone (vlc_thread_t *p_handle, void * (*entry) (void *), void *data,
               int priority)
{
    int err = ENOMEM;

    struct vlc_entry_data *entry_data = malloc (sizeof (*entry_data));
    if (entry_data == NULL)
        return ENOMEM;

    entry_data->func = entry;
    entry_data->data = data;
    vlc_sem_init (&entry_data->ready, 0);

    vlc_thread_t th = malloc(sizeof( *th ));
    if( th == NULL )
        goto error;

    if( DosCreateEventSem( NULL, &th->hevCancel, 0, FALSE ))
        goto error;
    if( DosCreateEventSem( NULL, &th->hevDone, 0, FALSE ))
        goto error;

    entry_data->hevCancel = th->hevCancel;
    entry_data->hevDone   = th->hevDone;
    th->tid = _beginthread( vlc_entry, NULL, 1024 * 1024, entry_data );
    if(( int )th->tid == -1 )
        goto error;

    *p_handle = th;

    if (priority)
        DosSetPriority( PRTYS_THREAD,
                        HIBYTE( priority ),
                        LOBYTE( priority ),
                        th->tid );

    /* Prevent cancellation until cancel_data is initialized. */
    /* XXX: This could be postponed to vlc_cancel() or avoided completely by
     * passing the "right" pointer to vlc_cancel_self(). */
    vlc_sem_wait (&entry_data->ready);
    vlc_sem_destroy (&entry_data->ready);
    free (entry_data);

    return 0;

error:
    if (th)
    {
        DosCloseEventSem (th->hevCancel);
        DosCloseEventSem (th->hevDone);
        free (th);
    }
    vlc_sem_destroy (&entry_data->ready);
    free (entry_data);
    return err;
}

void vlc_join (vlc_thread_t handle, void **result)
{
    ULONG rc;

    do
    {
        vlc_testcancel();
        rc = DosWaitEventSemEx( handle->hevDone, SEM_INDEFINITE_WAIT, TRUE );
    } while( rc == ERROR_INTERRUPT );

    assert (result == NULL); /* <- FIXME if ever needed */

    DosCloseEventSem( handle->hevCancel );
    DosCloseEventSem( handle->hevDone );

    free( handle );
}

void vlc_detach (vlc_thread_t handle)
{
    /* FIXME: handle->hevCancel and handle->hevDone leak */
    free( handle );
}

/*** Thread cancellation ***/

/* APC procedure for thread cancellation */
static void vlc_cancel_self (PVOID dummy)
{
    vlc_cancel_t *nfo = vlc_threadvar_get (cancel_key);

    if (likely(nfo != NULL))
        nfo->killed = true;

    (void)dummy;
}


void vlc_cancel (vlc_thread_t thread_id)
{
    DosPostEventSem( thread_id->hevCancel );
}

int vlc_savecancel (void)
{
    int state;

    vlc_cancel_t *nfo = vlc_threadvar_get (cancel_key);
    if (nfo == NULL)
        return false; /* Main thread - cannot be cancelled anyway */

    state = nfo->killable;
    nfo->killable = false;
    return state;
}

void vlc_restorecancel (int state)
{
    vlc_cancel_t *nfo = vlc_threadvar_get (cancel_key);
    assert (state == false || state == true);

    if (nfo == NULL)
        return; /* Main thread - cannot be cancelled anyway */

    assert (!nfo->killable);
    nfo->killable = state != 0;
}

void vlc_testcancel (void)
{
    vlc_cancel_t *nfo = vlc_threadvar_get (cancel_key);
    if (nfo == NULL)
        return; /* Main thread - cannot be cancelled anyway */

    /* This check is needed for the case that vlc_cancel() is followed by
     * vlc_testcancel() without any cancellation point as in signals handler
     * module */ 
    if( DosWaitEventSem( nfo->hevCancel, 0 ) == NO_ERROR )
        vlc_cancel_self( NULL );

    if (nfo->killable && nfo->killed)
    {
        for (vlc_cleanup_t *p = nfo->cleaners; p != NULL; p = p->next)
             p->proc (p->data);

        DosPostEventSem( nfo->hevDone );
        _endthread();
    }
}

void vlc_control_cancel (int cmd, ...)
{
    /* NOTE: This function only modifies thread-specific data, so there is no
     * need to lock anything. */
    va_list ap;

    vlc_cancel_t *nfo = vlc_threadvar_get (cancel_key);
    if (nfo == NULL)
        return; /* Main thread - cannot be cancelled anyway */

    va_start (ap, cmd);
    switch (cmd)
    {
        case VLC_CLEANUP_PUSH:
        {
            /* cleaner is a pointer to the caller stack, no need to allocate
             * and copy anything. As a nice side effect, this cannot fail. */
            vlc_cleanup_t *cleaner = va_arg (ap, vlc_cleanup_t *);
            cleaner->next = nfo->cleaners;
            nfo->cleaners = cleaner;
            break;
        }

        case VLC_CLEANUP_POP:
        {
            nfo->cleaners = nfo->cleaners->next;
            break;
        }
    }
    va_end (ap);
}


/*** Timers ***/
struct vlc_timer
{
    TID    tid;
    HEV    hev;
    HTIMER htimer;
    bool   quit;
    void (*func) (void *);
    void  *data;
};

static void vlc_timer_do (void *arg)
{
    struct vlc_timer *timer = arg;

    while (1)
    {
        ULONG count;

        DosWaitEventSem (timer->hev, SEM_INDEFINITE_WAIT);
        DosResetEventSem (timer->hev, &count);

        if (timer->quit)
            break;

        timer->func (timer->data);
    }
}

int vlc_timer_create (vlc_timer_t *id, void (*func) (void *), void *data)
{
    struct vlc_timer *timer = malloc (sizeof (*timer));

    if (timer == NULL)
        return ENOMEM;

    timer->func = func;
    timer->data = data;

    DosCreateEventSem (NULL, &timer->hev, DC_SEM_SHARED, FALSE);
    timer->htimer = NULLHANDLE;
    timer->quit = false;
    timer->tid  = _beginthread (vlc_timer_do, NULL, 1024 * 1024, timer);

    *id = timer;
    return 0;
}

void vlc_timer_destroy (vlc_timer_t timer)
{
    if (timer->htimer != NULLHANDLE)
    {
        DosStopTimer (timer->htimer);
        timer->quit = true;
        DosPostEventSem (timer->hev);
        DosWaitThread (&timer->tid, DCWW_WAIT);
        DosCloseEventSem (timer->hev);
    }

    free (timer);
}

void vlc_timer_schedule (vlc_timer_t timer, bool absolute,
                         mtime_t value, mtime_t interval)
{
    if (timer->htimer != NULLHANDLE)
    {
        DosStopTimer (timer->htimer);
        timer->htimer = NULLHANDLE;
    }

    if (value == 0)
        return; /* Disarm */

    if (absolute)
        value -= mdate ();
    value = (value + 999) / 1000;
    interval = (interval + 999) / 1000;

    if (DosStartTimer (interval, (HSEM) timer->hev, &timer->htimer))
        abort ();
}

unsigned vlc_timer_getoverrun (vlc_timer_t timer)
{
    (void)timer;
    return 0;
}
